module Images (Get, Pixel, Image, ImageGetter, getImage, getImageExt, printImage, printImageExt) where

import Data.Binary.Get		(runGet)
import IO					(IOMode (ReadMode))
import System.IO			(openBinaryFile)
import Data.ByteString.Lazy	(hGetContents)
import Maybe				(fromMaybe)
import Images.Types			(Get, Pixel, Image, ImageGetter)
import qualified Images.Bmp	(getImage)
import qualified Images.Png	(getImage)

defaultImageGetters :: [ImageGetter]
defaultImageGetters =
	[
		 ("bmp", Images.Bmp.getImage)
		,("png", Images.Png.getImage) -- This is a placeholder, PNG files are not currently supported
	]


getImage :: FilePath -> IO Image
getImage path = getImageExt path []


-- Figures out which "image getter" to use for the target file
getImageExt :: FilePath -> [ImageGetter] -> IO Image
getImageExt path imageGetters = do
	handle <- openBinaryFile path ReadMode
	contents <- hGetContents handle
	return (check (runGet imageGetter contents))
	where
		imageGetter	= fromMaybe undefined (lookup extension (defaultImageGetters ++ imageGetters))
		extension	= reverse (takeWhile (/= '.') (reverse path))

		check :: Image -> Image
		check image
			| width >= 0
				&& height >= 0
				&& width * height == length pixels	= image
			| otherwise 							= error ("Bad image: " ++ show image)
			where
				(width, height, pixels) = image


printImage :: FilePath -> IO ()
printImage path = printImageExt path []


printImageExt :: FilePath -> [ImageGetter] -> IO ()
printImageExt path imageGetters = do
	image <- getImageExt path imageGetters
	print (show image)


{-
	The remainder of this file is an attempt to support arbitrary file types.  After a lot of frustrating experimentation I couldn't get this to work.
	I suspect my OOP background has closed my mind a little and made it difficult to see how the type system really works in Haskell.
        I've come to the tentative conclusion that what I was trying to do here would be possible in Generic Haskell but I hope there is a simpler way
        that I've just missed.

	The idea was to have an extensible system for reading image files with some formats built in but leaving the possibility of additional formats to be plugged in later.
	I wanted to have a dictionary of file extension/"function" tuples where the function was capable of parsing files of that type and returning it without having to copy
	into some new canonical type.  For example, the types BmpFile and PngFile are structurally different but are both members of the class Image allowing this code to
	work with either through that classes methods.  As it is, I've reverted back to using a canonical Image type.

	The following code and comments are just various bits and pieces that I might find use for this in the future.

	TODO: Is there any way to alias "(Image a) => [(String, Get a)]" as ImageGetter?
		Using a class constraint in a type declaration is not permitted
		Using a newtype or data declaration forces the use of a constructor which is messy
		This question probably shows that I'm missing something important...

class (Show a) => Image a where
	width		:: a -> Int
	height		:: a -> Int
	pixels		:: a -> [Pixel]
	getImage	:: Get a

defaultImageGetters :: (Image a) => [(String, Get a)]
defaultImageGetters =
	[
		 ("bmp", getBmpFile)
		--,("png", getPngFile)
	]

-- TODO: Support files with no extension (e.g. peek at file type marker at start of file)
getImageExt :: (Image a) => FilePath -> [(String, Get a)] -> IO a
getImageExt path imageGetters = do
	handle <- openBinaryFile path ReadMode
	contents <- hGetContents handle
	return (runGet imageGetter contents)
	where
		imageGetter	= fromMaybe undefined (lookup extension (defaultImageGetters ++ imageGetters))
		extension	= reverse (takeWhile (/= '.') (reverse path))

getImage :: (Image a) => FilePath -> IO a
getImage path = getImageExt path []

printImageExt :: (Image a) => FilePath -> [(String, Get a)] -> IO ()
printImageExt path imageGetters = do
	image <- getImageExt path imageGetters
	print (show image)

--TODO: Another typing question, do I really have to conjure up a dummy type just so I can inform GHC what type the empty list really is?

data NilFile = CNilFile
	deriving (Show)

instance Image NilFile where
	width  _	= undefined
	height _	= undefined
	pixels _	= undefined

printImage :: FilePath -> IO ()
printImage path = printImageExt path ([] :: [(String, Get NilFile)])
-}
