module Images.Bmp (getImage) where

import Images.Types		(Get, Pixel, Image)
import Data.Binary.Get	(getWord16le, getWord32le, getWord8, skip, remaining)

-- BMP file format specification taken from MSDN:
-- http://msdn.microsoft.com/en-us/library/ms532311(VS.85).aspx

{-

  This is not intended to be a complete implementation of the BMP specification.
  Currently only 24 bit BMP files are supported and unusual header structures may cause a filure too
    although some effort is taken to skip newer header fields

  The Data.Binary.Get API makes this kind of processing very easy!

-}

data BmpFileHeader =
	BmpFileHeader
	{
		bfType :: Int,
		bfSize :: Int,
		bfReserved1 :: Int,
		bfReserved2 :: Int,
		bfOffBits :: Int
	}
	deriving (Show)

data BmpInfoHeader =
	BmpInfoHeader
	{
		biSize :: Int,
		biWidth :: Int,
		biHeight :: Int,
		biPlanes :: Int,
		biBitCount :: Int,
		biCompression :: Int,
		biSizeImage :: Int,
		biXPelsPerMeter :: Int,
		biYPelsPerMeter :: Int,
		biClrUsed :: Int,
		biClrImportant :: Int
	}
	deriving (Show)

type BmpPixel = Pixel

data BmpFile = CBmpFile BmpFileHeader BmpInfoHeader [BmpPixel]

instance Show BmpFile where
	show (CBmpFile bf bi ps) = show (bf, bi, length ps, foldr mix (0, 0, 0) ps)
		where
			-- Force every byte to be read fully (helped find some IO problems)
			mix (r1, g1, b1) (r2, g2, b2) = (r1 `avg` r2, g1 `avg` g2, b1 `avg` b2)
			x `avg` y = (x + y) `div` 2

getBmpFileHeader :: Get BmpFileHeader
getBmpFileHeader = do
		bfType <- getWord16le
		bfSize <- getWord32le
		bfReserved1 <- getWord16le
		bfReserved2 <- getWord16le
		bfOffBits <- getWord32le
		return BmpFileHeader
			{
				bfType = fromIntegral bfType,
				bfSize = fromIntegral bfSize,
				bfReserved1 = fromIntegral bfReserved1,
				bfReserved2 = fromIntegral bfReserved2,
				bfOffBits = fromIntegral bfOffBits
			}

getBmpInfoHeader :: Get BmpInfoHeader
getBmpInfoHeader = do
		biSize <- getWord32le
		biWidth <- getWord32le
		biHeight <- getWord32le
		biPlanes <- getWord16le
		biBitCount <- getWord16le
		biCompression <- getWord32le
		biSizeImage <- getWord32le
		biXPelsPerMeter <- getWord32le
		biYPelsPerMeter <- getWord32le
		biClrUsed <- getWord32le
		biClrImportant <- getWord32le
		return BmpInfoHeader
			{
				biSize = fromIntegral biSize,
				biWidth = fromIntegral biWidth,
				biHeight = fromIntegral biHeight,
				biPlanes = fromIntegral biPlanes,
				biBitCount = fromIntegral biBitCount,
				biCompression = fromIntegral biCompression,
				biSizeImage = fromIntegral biSizeImage,
				biXPelsPerMeter = fromIntegral biXPelsPerMeter,
				biYPelsPerMeter = fromIntegral biYPelsPerMeter,
				biClrUsed = fromIntegral biClrUsed,
				biClrImportant = fromIntegral biClrImportant
			}

getBmpPixels :: Int -> Int -> Int -> Get [BmpPixel]
getBmpPixels _ _ 0 = return []
getBmpPixels w 0 y = do
	-- Skip over row padding (each row is aligned to a 32-bit word boundary)
	skip (w - (4 * (w `div` 4)))
	getBmpPixels w w (y - 1)
getBmpPixels w x y = do
	remaining <- remaining
	-- Colour bytes stored in reverse order
	b <- if remaining < 3
		then fail ("Ran out of data at x:" ++ show x ++ ", y:" ++ show y)
		else getWord8
	g <- getWord8
	r <- getWord8
	remainder <- getBmpPixels w (x - 1) y
	return ((fromIntegral r, fromIntegral g, fromIntegral b) : remainder)

getBmpFile :: Get BmpFile
getBmpFile = do
	bf <- getBmpFileHeader
	bi <- getBmpInfoHeader
	if biBitCount bi /= 24
		then fail "Only 24 bit BMP files are supported."
		-- Skip over any remaining header fields (e.g. if V4 or V5 BMP)
		else skip (bfOffBits bf - 54)
	ps <- getBmpPixels (biWidth bi) (biWidth bi) (biHeight bi)
	return (CBmpFile bf bi ps)

getImage :: Get Image
getImage = do
	(CBmpFile _ bi ps) <- getBmpFile
	return (biWidth bi, biHeight bi, ps)
