module Main (main, displayByLSystem, displayByShapes) where

{-
	Some portions of this file, especially the OpenGL and execute functions are
	taken from the original LSystem module by Willem Heijltjes & Phil Wadler
	and originally based on OpenGL, GLFW, and the SOE library by Paul Liu and
	Duncan Coutts, using fragments of SOE's accompanying drawing modules.  Some
	of these have been modified further by Daniel Renshaw.
-}

import Main.Types		(Angle, LSystem, Shape, Distance, Turtle, Point (CPoint), Line (CLine), Bounds (CBounds), Command (..), Pen (Colour, Inkless))
import Main.Utilities	(left, right, black, white, pointToTuple, originPointF, originPointI, pointExpand, splitBy)
import Main.LSystems	(getLSystem)
import Main.Shapes		(getShapes)
import Images			(Image, getImage, Pixel)
import System.Random	(getStdRandom, StdGen, randomRs, split)
import System			(getArgs)
import Test.QuickCheck	(quickCheck, verboseCheck)
import Pictures			(openWindow, drawInWindow, closeWindow, Graphic (Graphic), Window, getKey)
import qualified Graphics.Rendering.OpenGL as GL

{-
  To use this application, it is recommended that it be compiled first; you may have stack overflow problems if run interactively and processing large images

  To compile:  ghc --make -O3 MainExe.hs

  To run, using either of the following command line forms:

  ./MainExe "byLSystem" "SampleImages/HaskellCurrySm.bmp" "hexagonalGosper" 5 +RTS -K50M

  ./MainExe "byShapes" "SampleImages/HaskellCurrySm.bmp" "lambdaL,sigmaU" 2 +RTS -K50M

  In the first form, the parameters are: <Mode> <InputImage> <LSystemName> <NoIterations>

  In the second form, the parameters are: <Mode> <InputImage> <ShapesToUse> <InterShapeGapSize>

  In both cases the +RTS and -K50M change the stack space size and may or may not be needed.  These are more likely to be needed when processing large images.

  In this distribution, the only supported modes are "byLSystem" and "byShapes". The only useful L-Systems are "hexagonalGosper" and "hilbert" (other L-Systems
    are included, but they do not cover enough space to be useful). The only supported shapes are "lambdaL" and "sigmaU".

  Any number of shapes can be provided, comma separated in <ShapesToUse>

  Additional shapes and L-Systems can be added to the Main/LSystems.hs and Main/Shapes.hs files respectively.  You'll need to alter the dictionary near
    the top of each file to register the new shapes in each case.

  Apologies if the code layout seems a bit messy - I wrote most of this using tabs rather than spaces.  Setting the tab length to 4 characters might sort it out.

  There are many things I would choose to improve if given more time and motivation, a selection follow:

    - Performance and esp. stack space overflow
      - May be related to use of !!, could this be improved by using the "array" module?
      - May be related to reading bytes in file order when it may be better to read them in image order (if possible)
    - Figure out if better polymorphic support for "Point a" can be offered instead of the many fixed "Point Int" and "Point Float" instances
    - Handle bad command lines more gracefuly
    - Add more error handling in general!
    - Support image files with no extension (e.g. peek at file type marker at start of file)
    - Allow plugged in image getters to over-ride built in ones for the same type
    - Figure out the generic version of the image getters stuff that does away with the canonical Image type
-}


{- OPEN-GL functions -}

-- The following functions come from LSystem.hs but I've altered them a bit
penToColor3 :: Pen -> GL.Color3 Float
penToColor3 (Colour r g b)	= GL.Color3 r g b
penToColor3 Inkless			= error "penToColor3: inkless"

displayWindow :: [Line Float] -> IO ()
displayWindow lines	= do
	window <- openWindow "INF1A programming competition, entry by Daniel Renshaw, s0830457" (pointToTuple dimensions)
	drawInWindow window (graphicBackground dimensions)
	drawInWindow window (graphicLines (rescale (CBounds originPointF (pointExpand dimensions)) lines))
	spaceClose window
	where
		(CBounds (CPoint minX minY) (CPoint maxX maxY)) = calcBounds lines
		xDelta = maxX - minX
		yDelta = maxY - minY
		-- Set the dimensions of the window proportional to the image
		dimensions = if xDelta >= yDelta
			then CPoint 1152 (round (yDelta * (1152 / xDelta)))
			else CPoint (round (xDelta * (864 / yDelta))) 864

toVertex :: Point Float -> IO ()
toVertex (CPoint x y)	= GL.vertex (GL.Vertex3 x y 0)

graphicBackground :: Point Int -> Graphic
graphicBackground canvas	= Graphic $
	GL.color (penToColor3 white) >>
	GL.renderPrimitive GL.Polygon (sequence_ (map toVertex [originPointF, CPoint x 0, CPoint x y, CPoint 0 y]))
	where
		CPoint x y	= pointExpand canvas

graphicLines :: [Line Float] -> Graphic
graphicLines lines	= Graphic (sequence_ (map f lines))
	where
		f (CLine pen start end)	=
			GL.color (penToColor3 pen) >>
			GL.renderPrimitive GL.LineStrip (toVertex start >> toVertex end)

spaceClose :: Window -> IO ()
spaceClose window	= do
	key <- getKey window
	if (key == ' ' || key == '\x0')
		then closeWindow window
		else spaceClose window


{- Command functions -}

-- Other than switching to "Pnt" to "Point Float", this is a direct copy of the "execute" function int eh original LSystem.hs
execute :: Command -> [Line Float]
execute c	= lines
	where
		(lines, turtle)	= f c (black, 270, originPointF)

		f :: Command -> Turtle -> ([Line Float], Turtle)
		f (c :#: d) turtle					= (clines ++ dlines, dturtle)
			where
				(clines, cturtle)	= f c turtle
				(dlines, dturtle)	= f d cturtle

		f (Branch c) turtle					= (clines, turtle)
			where
				(clines, cturtle) = f c turtle

		f (Go distance) (pen, angle, point)	= (if pen == Inkless
			then []
			else [CLine pen point end]
			, (pen, angle, end))
			where
				end = point + (CPoint distance distance) * polar angle

		f (Turn delta) (pen, angle, point)	=  ([], (pen, angle - delta, point))
		f (GrabPen new) (old, angle, point)	=  ([], (new, angle, point))
		f (Sit) turtle						=  ([], turtle)

		polar :: Angle -> Point Float
		polar angle	= CPoint (cos radians) (sin radians)
			where
				radians	= angle * 2 * pi / 360


-- NOTE: This is a polymorphic version of the fixed type versions in Utilities, it is only used in the following function right now
originPoint :: (Real a) => Point a
originPoint = CPoint 0 0


calcBounds :: (Real a) => [Line a] -> Bounds a
calcBounds = foldr calcBounds' (CBounds originPoint originPoint)
	where
		calcBounds' :: (Real a) => Line a -> Bounds a -> Bounds a
		calcBounds' (CLine _ (CPoint x0 y0) (CPoint x1 y1)) (CBounds (CPoint minX minY) (CPoint maxX maxY))
			= (CBounds (CPoint (minX `min` x0 `min` x1) (minY `min` y0 `min` y1)) (CPoint (maxX `max` x0 `max` x1) (maxY `max` y0 `max` y1)))


rescale :: Bounds Float -> [Line Float] -> [Line Float]
rescale (CBounds (CPoint dMinX dMinY) (CPoint dMaxX dMaxY)) lines = map rescale' lines
	where
		(CBounds (CPoint sMinX sMinY) (CPoint sMaxX sMaxY)) = calcBounds lines

		rescale' :: Line Float -> Line Float
		rescale' (CLine pen (CPoint x0 y0) (CPoint x1 y1)) = CLine pen start end
			where
				xScale	= (dMaxX - dMinX) / (sMaxX - sMinX)
				yScale	= (dMaxY - dMinY) / (sMaxY - sMinY)
				start	= (CPoint (calc x0 sMinX xScale dMinX) (calc y0 sMinY yScale dMinY))
				end		= (CPoint (calc x1 sMinX xScale dMinX) (calc y1 sMinY yScale dMinY))

				calc :: Float -> Float -> Float -> Float -> Float
				calc v sMin scale dMin	= ((v - sMin) * scale) + dMin


{- Line and Point functions -}

-- See http://en.wikipedia.org/wiki/Bresenham's_line_algorithm for details (although I've enhanced it to work with negative X values)
-- NOTE: This was originally a little more general but I've had to compromise the design in order to avoid a problem of scaling between "Line Float" and "Line Int" and back
bresenham :: Line Float -> (Line Float, [Point Int])
bresenham (CLine pen (CPoint x0 y0) (CPoint x1 y1))	= ((CLine pen (CPoint x0 y0) (CPoint x1 y1)), bresenham' (CPoint x0' y0') 0)
	where
		(x0r, y0r, x1r, y1r)	= (round x0, round y0, round x1, round y1)
		(x0', y0', x1', y1')	= if steep
			then (y0r, x0r, y1r, x1r)
			else (x0r, y0r, x1r, y1r)

		steep					= abs (y1r - y0r) > abs (x1r - x0r)
		deltaErr				= fromIntegral (abs (y1' - y0')) / fromIntegral (abs (x1' - x0'))
		xStep					= if x0' < x1' then 1 else (-1)
		yStep					= if y0' < y1' then 1 else (-1)

		bresenham' :: Point Int -> Float -> [Point Int]
		bresenham' (CPoint x y) err
			| x == x1'			= [point]
			| otherwise			= point : bresenham' (CPoint nextX nextY) nextErr
			where
				point	= if steep then (CPoint y x) else (CPoint x y)
				err'	= err + deltaErr
				nextX	= x + xStep
				nextY	= if err' >= 0.5 then y + yStep else y
				nextErr	= if err' >= 0.5 then err' - 1  else err'


colourLines :: Image -> [(Line Float, [Point Int])] -> [Line Float]
colourLines (width, height, pixels) = map colourLine
	where
		colourLine :: (Line Float, [Point Int]) -> Line Float
		colourLine (CLine _ start end, points)	= CLine (Colour ra ga ba) start end
			where
				-- Calculate the average colour of traversed pixels
				pointCount		= fromIntegral (length points)
				(rs, gs, bs)	= foldr addPixels (0, 0, 0) (map getPixel points)
				(ra, ga, ba)	= (convert rs, convert gs, convert bs)
				convert	colour	= (fromIntegral colour / pointCount) / 255

				getPixel :: Point Int -> Pixel
				-- BMP rows are in reverse order, hence the need for "height - y -1"
				getPixel (CPoint x y)	= pixels!!(((height - y - 1) * width) + x)

				addPixels :: (Int, Int, Int) -> (Int, Int, Int) -> (Int, Int, Int)
				addPixels (r0, g0, b0) (r1, g1, b1)	= (r0 + r1, g0 + g1, b0 + b1)


{- Main display functions -}

-- Produces a coloured version of each line, using the average colour of the pixels traversed
-- by the line when overlaid on an image
processLines :: Image -> [Line Float] -> IO [Line Float]
processLines (width, height, pixels) lines = return colouredLines
	where
		imageDimension		= pointExpand (CPoint (width - 1) (height - 1))
		-- Rescale the lines to the dimensions of the image
		rescaledToImage		= rescale (CBounds originPointF imageDimension) lines
		-- Use the Bresenham line drawing algorithm to find the pixels traversed by each line
		linePoints			= map bresenham rescaledToImage
		-- Set the colour of the line to the average colour of the pixels it traverses
		colouredLines		= colourLines (width, height, pixels) linePoints


-- Given a random number, selects an element from the list
randomGet :: Float -> [a] -> a
randomGet r list = list !! round (r * fromIntegral (length list - 1))


-- Produces a turtle command where each pixel in the image is represented by a shape drawn
-- randomly from the shapes list and twisted by a random angle leaving the gapSize between
-- each shape
drawShapes :: Image -> [Shape] -> [Float] -> Distance -> IO Command
drawShapes (width, height, pixels) shapes randoms gapSize
	= return (GrabPen Inkless :#: right 90 :#: pixel width height pixels randoms)
	where
		baseSize, shapeSize, boxSize, rowSize :: Distance
		baseSize	= 1
		shapeSize	= 10 * baseSize
		boxSize		= shapeSize + gapSize
		rowSize		= boxSize * fromIntegral width

		pixel :: Int -> Int -> [Pixel] -> [Float] -> Command
		pixel 0 1 [] _			=	Sit
		pixel 0 y ps rs			=		left  180 :#: Go rowSize
									:#: right  90 :#: Go boxSize
									:#: right  90 :#: pixel width (y - 1) ps rs

		pixel x y (p:ps) (r:rs)	=		Branch
										(
											twist r angles
											:#: grabPen p
											:#: command
										)
									:#: Go boxSize :#: pixel (x - 1) y ps rs
			where
				(command, angles)	= shape r

		pixel x y [] _			= error ("Ran out of pixels at x:" ++ show x ++ ", y:" ++ show y)

		-- drawBox used for debugging purposes only
		drawBox :: Command
		drawBox =		Go boxSize :#: left 90 :#: Go boxSize :#: left 90
					:#: Go boxSize :#: left 90 :#: Go boxSize :#: left 90

		shape :: Float -> Shape
		shape r = randomGet r shapes

		twist :: Float -> [Angle] -> Command
		twist r	angles =	Go (boxSize / 2)	:#: left  90
						:#: Go (boxSize / 2)	:#: left  (randomGet r angles)
						:#: Go (shapeSize / 2)	:#: right 90
						:#: Go (shapeSize / 2)	:#: right 90

		grabPen :: Pixel -> Command
		grabPen (r, g, b) = GrabPen (Colour (convert r) (convert g) (convert b))
			where
				convert :: Int -> Float
				convert = (/255).fromIntegral

-- Executes an L-System, overlays it on an image and colours the lines accordingly
displayByLSystem :: FilePath -> LSystem -> Int -> IO ()
displayByLSystem path lSystem iterations = do
	image	<- getImage path
	result	<- processLines image lines
	displayWindow result
	where
		pattern	= lSystem iterations
		lines	= execute pattern


-- Redraws an image with each pixel replaced by a Turtle command "shape"
displayByShapes :: FilePath -> [Shape] -> Distance -> IO ()
displayByShapes path shapes gapSize = do
	image	<- getImage path
	randoms <- getStdRandom generator
	result	<- drawShapes image shapes randoms gapSize
	displayWindow (execute result)
	where
		generator :: StdGen -> ([Float], StdGen)
		generator r = (randomRs (0.0, 1.0) r1, r2)
			where
				(r1, r2) = split r


-- Provides a very basic mechanism for executing the program after compilation from the command line
main :: IO ()
main = do
	args <- getArgs
	case args!!0 of
		"byLSystem"	-> displayByLSystem (args!!1) (getLSystem (args!!2)) (read (args!!3))
		"byShapes"	-> displayByShapes (args!!1) (getShapes (splitBy ',' (args!!2))) (read (args!!3))
